"""Query module.

See https://docs.arangodb.com/Aql/Basics.html for details

"""

from functools import wraps
from itertools import izip, imap

from collections import defaultdict, OrderedDict

from . import meta, util, cursor


class Expression(object):

    """Everything is an expression.

    The only thing to remenber is, that you should only directly yield a
    subexpression, if it implements :py:meth:`.Expression._get_term`, if that is
    not the case, just yield from it.

    """

    def __iter__(self):
        """Iterate over all terms and return a tuple of query string part and related bind params."""

        yield self

    def _get_params(self, index):           # pylint: disable=R0201
        """:returns: indexed params"""

        return {}

    def _get_term(self, index):
        raise NotImplementedError("If this expression should have a term, you have to implement it")

    def assemble(self):
        # hold index for each type and instance
        # we use an OrderedDict here because testing for list containment
        # is not working with overloading compare operators
        expressions = defaultdict(OrderedDict)

        def register_expr(expr):
            types = expressions[type(expr)]

            if expr not in types:
                types[expr] = len(types)

            return types[expr]

        for expr in self:
            index = register_expr(expr)

            yield expr._get_term(index), expr._get_params(index)            # pylint: disable=w0212

    def query(self):
        """Create a query with its bind params assembled."""

        terms, binds = izip(*list(self.assemble()))

        joined_params = {}
        for params in binds:
            joined_params.update(params)

        return ''.join(terms), joined_params

    def __eq__(self, other):
        return Operator(self, other, _EQ)

    def __lt__(self, other):
        return Operator(self, other, _LT)

    def __le__(self, other):
        return Operator(self, other, _LE)

    def __gt__(self, other):
        return Operator(self, other, _GT)

    def __ge__(self, other):
        return Operator(self, other, _GE)


class Param(object):

    """Create a bind parameter."""

    def __init__(self, name):
        self.name = name

    def __call__(self, index, declare=True):
        return "{declare}{name}_{index}"\
            .format(declare=declare and "@" or "", name=self.name, index=index)


class Value(Expression):

    """Inject values into a Query.

    A :py:class:`.Value` will create a bind parameter.
    """

    param = Param('value')

    def __init__(self, value):
        self.value = value

    def _get_term(self, index):
        return self.param(index)

    def _get_params(self, index):
        return {
            self.param(index, False): self.value
        }

    def __repr__(self):
        return "<{0.__class__.__name__}: {0.value}>".format(self)

    @classmethod
    def iter_fixed(cls, sequence):
        """:returns: a generator which is wrapping none-expressions into :py:class:`.Value`"""

        # TODO do we need an iterator for non-sequences?
        for value in sequence:
            yield cls.fix(value)

    @classmethod
    def fix(cls, value):
        """If value is not an expression make it a :py:class:`.Value`."""

        if not isinstance(value, Expression):
            return cls(value)

        return value


def iter_fix_value(wrapped):

    @wraps(wrapped)
    def decorator(*args, **kwargs):
        return Value.iter_fixed(wrapped(*args, **kwargs))

    return decorator


class Term(Expression):
    term = ''

    def __init__(self, term):
        self.term = str(term)

    def __iter__(self):
        yield self

    def _get_params(self, index):
        return {}

    def _get_term(self, index):
        return self.term

    def __repr__(self):
        return "<{0.__class__.__name__}: {0.term}>".format(self)


# some semantics
SPACE = Term(" ")
LPAR = Term("(")
RPAR = Term(")")


class KeyWord(Term):
    pass


FOR = KeyWord("FOR")
RETURN = KeyWord("RETURN")
FILTER = KeyWord("FILTER")
SORT = KeyWord("SORT")
LIMIT = KeyWord("LIMIT")
LET = KeyWord("LET")
COLLECT = KeyWord("COLLECT")
INSERT = KeyWord("INSERT")
UPDATE = KeyWord("UPDATE")
REPLACE = KeyWord("REPLACE")
REMOVE = KeyWord("REMOVE")
WITH = KeyWord("WITH")
ASC = KeyWord("ASC")
DESC = KeyWord("DESC")
IN = KeyWord("IN")
INTO = KeyWord("INTO")
NOT = KeyWord("NOT")
AND = KeyWord("AND")
OR = KeyWord("OR")
NULL = KeyWord("NULL")
TRUE = KeyWord("TRUE")
FALSE = KeyWord("FALSE")


class Alias(Expression):
    def __init__(self, name):
        self.__name__ = name

    def _get_term(self, index):
        return self.__name__

    def __getattr__(self, name):
        attr = AliasAttr(self, name)

        self.__dict__[name] = attr
        return attr

    def __repr__(self):
        return "<{0.__class__.__name__}: {0.__name__}>".format(self)


class AliasAttr(Alias):
    def __init__(self, parent, name):
        super(AliasAttr, self).__init__(name)
        self.__parent__ = parent

    def __repr__(self):
        return "<{0.__class__.__name__}: {0.__parent__.__name__}.{0.__name__}>".format(self)

    def _get_term(self, index):
        parent_term = self.__parent__._get_term(index)          # pylint: disable=W0212

        return "{}.`{}`".format(parent_term, self.__name__)


# pylint: disable=W0223

class Chain(Expression):
    # sep may be a chain
    sep = None

    def __init__(self, exprs, **kwargs):
        """
        :param exprs: a sequence of expressions.
        :param sep: an optional separator
        """

        sep = kwargs.get("sep")
        if sep is not None:

            # make sep a Term
            if not isinstance(sep, Expression):
                sep = Term(sep)

            self.sep = sep

        self.exprs = list(exprs)

    def append(self, expr):
        """Add an expression to the end of the list."""

        self.exprs.append(expr)

    def extend(self, exprs):
        """Add an list of expressions to the end of the list."""

        self.exprs.extend(exprs)

    def __len__(self):
        return len(self.exprs)

    @iter_fix_value
    def __iter__(self):
        c = len(self.exprs)

        for i, expr in enumerate(Value.iter_fixed(self.exprs)):
            if expr is not None:
                for subexpr in expr:
                    yield subexpr

            if i < c - 1 and self.sep is not None:
                for sep_expr in self.sep:
                    yield sep_expr

    def __repr__(self):
        return "<{0.__class__.__name__}: {list}>"\
            .format(self, list=" ".join(imap(repr, self.exprs)))


class List(Chain):
    sep = Term(', ')


class And(List):
    sep = Chain((SPACE, AND, SPACE))

    def __init__(self, *exprs):
        super(And, self).__init__(exprs)


class Or(List):
    sep = Chain((SPACE, OR, SPACE))

    def __init__(self, *exprs):
        super(Or, self).__init__(exprs)


class Filter(And):

    def __iter__(self):
        yield FILTER
        yield SPACE

        for expr in super(Filter, self).__iter__():
            yield expr


class In(Expression):
    def __init__(self, alias, expr):
        self.alias_expr = alias

        # might be a value or expression
        self.expr = Value.fix(expr)

    def __iter__(self):
        for expr in self.alias_expr:
            yield expr

        yield SPACE
        yield IN
        yield SPACE

        for expr in self.expr:
            yield expr


class Function(List):

    def __init__(self, *exprs):
        super(Function, self).__init__(exprs)

    def __iter__(self):
        yield Term(self.name)
        yield LPAR

        for expr in super(Function, self).__iter__():
            yield expr

        yield RPAR

    @util.classproperty
    def name(cls):          # pylint: disable=E0213
        return cls.__name__


# See https://docs.arangodb.com/Aql/Functions.html
# pylama:ignore=E302,E701,C0321
# TYPES
class TO_BOOL(Function): pass
class TO_NUMBER(Function): pass
class TO_STRING(Function): pass
class TO_ARRAY(Function): pass
class TO_LIST(Function): pass
class IS_NULL(Function): pass
class IS_BOOL(Function): pass
class IS_NUMBER(Function): pass
class IS_STRING(Function): pass
class IS_ARRAY(Function): pass
class IS_LIST(Function): pass
class IS_OBJECT(Function): pass
class IS_DOCUMENT(Function): pass

# STRING
class CONCAT(Function): pass
class CONCAT_SEPARATOR(Function): pass
class CHAR_LENGTH(Function): pass
class LOWER(Function): pass
class UPPER(Function): pass
class SUBSTITUTE(Function): pass
class SUBSTRING(Function): pass
class LEFT(Function): pass
class RIGHT(Function): pass
class TRIM(Function): pass
class LTRIM(Function): pass
class RTRIM(Function): pass
class SPLIT(Function): pass
class REVERSE(Function): pass
class CONTAINS(Function): pass
class FIND_FIRST(Function): pass
class fIND_LAST(Function): pass
class LIKE(Function): pass
class MD5(Function): pass
class SHA1(Function): pass
class RANDOM_TOKEN(Function): pass

# NUMERIC
class FLOOR(Function): pass
class CEIL(Function): pass
class ROUND(Function): pass
class ABS(Function): pass
class SQRT(Function): pass
class RAND(Function): pass

# DATE
class DATE_TIMESTAMP(Function): pass
class DATE_ISO8601(Function): pass
class DATE_DAYOFWEEK(Function): pass
class DATE_YEAR(Function): pass
class DATE_MONTH(Function): pass
class DATE_DAY(Function): pass
class DATE_HOUR(Function): pass
class DATE_MINUTE(Function): pass
class DATE_SECOND(Function): pass
class DATE_MILLISECOND(Function): pass
class DATE_NOW(Function): pass

# ARRAY
class LENGTH(Function): pass
class FLATTEN(Function): pass
class MIN(Function): pass
class MAX(Function): pass
class AVERAGE(Function): pass
class SUM(Function): pass
class MEDIAN(Function): pass
class PERCENTILE(Function): pass
class VARIANCE_POPULATION(Function): pass
class STDDEV_POPULATION(Function): pass
class SDTDEV_SAMPLE(Function): pass
# class REVERSE(Function): pass
class FIRST(Function): pass
class LAST(Function): pass
class NTH(Function): pass
class POSITION(Function): pass
class SLICE(Function): pass
class UNIQUE(Function): pass
class UNION(Function): pass
class UNION_DISTINCT(Function): pass
class MINUS(Function): pass
class INTERSECTION(Function): pass
class APPEND(Function): pass
class PUSH(Function): pass
class UNSHIFT(Function): pass
class POP(Function): pass
class SHIFT(Function): pass
class REMOVE_VALUE(Function): pass
class REMOVE_VALUES(Function): pass
class REMOVE_NTH(Function): pass

# OBJECT/DOCUMENT
class MATCHES(Function): pass
class MERGE(Function): pass
class MERGE_RECURSIVE(Function): pass
class TRANSLATE(Function): pass
class HAS(Function): pass
class ATTRIBUTES(Function): pass
class VALUES(Function): pass
class ZIP(Function): pass
class UNSET(Function): pass
class KEEP(Function): pass
class PARSE_IDENTIFIER(Function): pass

# GEO
class NEAR(Function): pass
class WITHIN(Function): pass
class WITHIN_RECTANGLE(Function): pass
class IS_IN_POLYGON(Function): pass

# FULLTEXT
class FULLTEXT(Function): pass

# GRAPH
class EDGES(Function): pass
class NEIGHBORS(Function): pass
class TRAVERSAL(Function): pass
class TRAVERSAL_TREE(Function): pass
class SHORTEST_PATH(Function): pass
class PATHS(Function): pass

# MISC
class NOT_NULL(Function): pass
class FIRST_LIST(Function): pass
class FIRST_DOCUMENT(Function): pass
class COLLECTIONS(Function): pass
class CURRENT_USER(Function): pass
class DOCUMENT(Function): pass
class SKIPLIST(Function): pass
class CALL(Function): pass
class APPLY(Function): pass


_EQ = Term("==")
_LT = Term("<")
_LE = Term("<=")
_GT = Term(">")
_GE = Term(">=")
_NE = Term("!=")
_IN = Term("IN")
_NOT_IN = Term("NOT IN")

_AND = Term("AND")
_OR = Term("OR")
_NOT = Term("NOT")


class Operator(Chain):
    sep = SPACE
    op = _EQ

    def __init__(self, a, b, op=None):
        self.a = a
        self.b = b

        if op is not None:
            self.op = op

        super(Operator, self).__init__((Value.fix(self.a), self.op, Value.fix(self.b)))


class Let(Expression):
    def __init__(self, alias, expr):
        self.alias_expr = alias

        self.expr = Value.fix(expr)

    def __iter__(self):
        yield LET
        yield SPACE

        for expr in self.alias_expr:
            yield expr

        yield SPACE
        yield Term("=")
        yield SPACE

        for expr in self.expr:
            yield expr


class Action(Expression):
    """A Kind of mandatory command for a query to perform.

    RETURN, REMOVE, INSERT, UPDATE, REPLACE
    """


class Return(Action):
    def __init__(self, alias_or_object):
        self.alias = alias_or_object

    def __iter__(self):
        yield RETURN
        yield SPACE

        for expr in self.alias:
            yield expr


class SortCriteria(Expression):
    term = ASC

    def __init__(self, alias):
        self.alias_expr = alias

    def __iter__(self):
        for expr in self.alias_expr:
            yield expr

        yield SPACE
        yield self.term


class Asc(SortCriteria):
    pass


class Desc(SortCriteria):
    term = DESC


class Sort(List):
    def __init__(self, *exprs):
        super(Sort, self).__init__(exprs)

    def __iter__(self):
        yield SORT
        yield SPACE

        for expr in super(Sort, self).__iter__():
            yield expr


class Limit(Expression):
    def __init__(self, offset_count, count=None):
        if count is None:
            self.count, self.offset = offset_count, None

        else:
            self.count, self.offset = count, offset_count

    def __iter__(self):
        yield LIMIT
        yield SPACE

        if self.offset is None:
            yield Term(self.count)

        else:
            for expr in List((Term(self.offset), Term(self.count))):
                yield expr


class Collection(Expression):
    param = Param('@c')

    def __init__(self, collection):
        self.collection = getattr(collection, "__collection_name__", collection)

    def _get_params(self, index):
        return {
            self.param(index, False): self.collection
        }

    def _get_term(self, index):
        return self.param(index)

    def __repr__(self):
        return "<{0.__class__.__name__}: {0.collection}>".format(self)


class For(Expression):
    def __init__(self, alias, from_list):
        self.alias(alias)
        self.from_list(from_list)

    def __iter__(self):
        yield FOR
        yield SPACE

        for expr in self.alias_expr:
            yield expr

        yield SPACE
        yield IN
        yield SPACE

        for expr in self.list_expr:
            yield expr

    def alias(self, alias):
        self.alias_expr = alias

        return self

    def from_list(self, from_list):
        self.list_expr = from_list

        return self


class QueryBase(Expression):

    __metaclass__ = meta.MetaQueryBase


class Query(QueryBase):

    """A query will join into an arango query plus bind params."""

    def __init__(self, alias, from_list,
                 action=None, filter=None, sort=None, limit=None):         # pylint: disable=W0622
        self.for_exprs = [For(alias, from_list)]

        self.action_expr = action

        self.filter_expr = self._get_filter(filter)

        self.sort_expr = self._get_sort(sort)

        self.limit_expr = limit

    @staticmethod
    def _get_filter(filter):            # pylint: disable=W0622
        filter_expr = Filter()
        if isinstance(filter, Filter):
            filter_expr.extend(filter.exprs)

        elif isinstance(filter, (list, tuple)):
            filter_expr.extend(filter)

        elif filter is not None:
            filter_expr.append(filter)

        return filter_expr

    @staticmethod
    def _get_sort(sort):
        sort_expr = Sort()
        if isinstance(sort, Sort):
            sort_expr.extend(Sort.exprs)

        elif isinstance(sort, (list, tuple)):
            sort_expr.extend(sort)

        elif sort is not None:
            sort_expr.append(sort)

        return sort_expr

    def __iter__(self):
        for for_expr in self.for_exprs:
            for expr in for_expr:
                yield expr
        if len(self.filter_expr):
            yield SPACE
            for expr in self.filter_expr:
                yield expr

        if len(self.sort_expr):
            yield SPACE
            for expr in self.sort_expr:
                yield expr

        if self.limit_expr is not None:
            yield SPACE
            for expr in self.limit_expr:
                yield expr

        yield SPACE
        for expr in self.action_expr:
            yield expr

    def filter(self, *filters):
        """Adds a filter expr."""

        self.filter_expr.extend(filters)

        return self

    def join(self, alias, from_list):
        self.for_exprs.append(For(alias, from_list))

        return self

    def action(self, action):
        """Replace the action of that query."""

        if isinstance(action, Action):
            self.action_expr = action

        else:
            self.action_expr = Return(action)

        return self

    def sort(self, *criteria):
        self.sort_expr.extend(criteria)

        return self

    def limit(self, *args):
        """Set a limit to the query."""

        self.limit_expr = Limit(*args)

        return self

    def validate(self):
        """Will ask the server to parse the query without executing it."""

        query, _ = self.query()

        return self.__class__.api.parse(query)

    @property
    def cursor(self):
        """Return a cursor for this query, ready to iterate."""

        return cursor.Cursor(*self.query())



"""ArangoDB api."""


from functools import wraps, partial
from itertools import imap, chain

import requests
import requests.adapters

from . import exc

import logging

LOG = logging.getLogger(__name__)


def json_result():
    """Decorate an arango response call to extract json and perform error handling."""

    def decorator(func):

        @wraps(func)
        def wrapped(*args, **kwargs):
            response = func(*args, **kwargs)

            headers = dict(response.headers.lower_items())
            if headers.get('content-type', '').startswith('application/json'):
                json_content = response.json()

                # inspect response
                if "error" in json_content and json_content.get('error', False):
                    # create a polymorphic exception just by errorNum
                    code = json_content.get('code')
                    num = json_content.get('errorNum')
                    message = json_content.get('errorMessage')

                    raise exc.ApiError(
                        code=code,
                        num=num,
                        message=message,
                        func=func,
                        args=args,
                        kwargs=kwargs,
                    )

                # no error
                return json_content

            raise exc.ContentTypeException("No json content-type", response)

        return wrapped

    return decorator


class Client(object):

    """A client for arangodb server."""

    def __init__(self, database, endpoint="http://localhost:8529", session=None):
        # default database
        self.database = database

        self.endpoint = endpoint.rstrip("/")

        # may use an external session
        if session is None:
            adapter = requests.adapters.HTTPAdapter(
                pool_maxsize=100,
                pool_connections=100
            )

            session = requests.Session()
            session.mount(self.endpoint, adapter)

        self.session = session

        # arango specific api
        self.collections = Collections(self.api(self.database, 'collection'))
        self.documents = Documents(self.api(self.database, 'document'))
        self.edges = Edges(self.api(self.database, 'edge'))
        self.cursors = Cursors(self.api(self.database, 'cursor'))
        self.graphs = Graphs(self.api(self.database, 'gharial'))
        self.indexes = Indexes(self.api(self.database, 'index'))
        self.queries = Queries(self.api(self.database, 'query'))

    def url(self, *path):
        """Return a full url to the arangodb server."""

        return '/'.join(imap(str, chain((self.endpoint, ), path)))

    @json_result()
    def get(self, *path, **kwargs):
        return self.session.get(self.url(*path), **kwargs)

    @json_result()
    def post(self, *path, **kwargs):
        return self.session.post(self.url(*path), **kwargs)

    @json_result()
    def put(self, *path, **kwargs):
        return self.session.put(self.url(*path), **kwargs)

    @json_result()
    def patch(self, *path, **kwargs):
        return self.session.patch(self.url(*path), **kwargs)

    def head(self, *path, **kwargs):
        return self.session.head(self.url(*path), **kwargs)

    @json_result()
    def delete(self, *path, **kwargs):
        return self.session.delete(self.url(*path), **kwargs)

    def api(self, database, *path, **kwargs):
        """Just expose the HTTP methods to this session, by partially pre binding the path."""

        if database is None:
            prefix = ('_api', )

        else:
            prefix = ('_db', database, '_api')

        return ApiProxy(self, *chain(prefix, path), **kwargs)


class SystemClient(Client):

    """A client to the system database of an arangodb server."""

    def __init__(self, endpoint="http://localhost:8529", session=None):
        super(SystemClient, self).__init__(None, endpoint=endpoint, session=session)

        # database api is only allowed for system database
        self.databases = Databases(self.api(None, 'database'))

    def create_database(self, database):
        """Just create the actual database if not exists."""

        if database not in self.databases.databases:
            self.databases.create(database)


class ApiProxy(object):

    """A Proxy to the session, partially preselect parts of the url and request parameter."""

    def __init__(self, session, *path, **kwargs):
        # wrap the session and preselect api
        for method in ('get', 'post', 'put', 'patch', 'delete', 'head'):
            setattr(self, method, partial(getattr(session, method), *path, **kwargs))


class Api(object):
    def __init__(self, api_proxy):
        self.api = api_proxy


class Databases(Api):
    """Database stuff."""

    @property
    def databases(self):
        return self.api.get('user')['result']

    def create(self, name, *users):
        """Create a database."""

        data = dict(name=name)

        if users:
            # TODO validation of users
            data['users'] = users

        return self.api.post(json=data).get('result', False)

    def drop(self, name):
        """Drop a database."""

        return self.api.delete(name).get('result', False)


DOCUMENT_COLLECTION = 2
EDGE_COLLECTION = 3


class Collections(Api):
    """Collection stuff."""

    def create(self, name, **kwargs):
        """Create a collection."""

        body = dict(name=name, **kwargs)

        return self.api.post(json=body)

    def get(self, *name, **kwargs):
        """Get one or all collection/s."""

        params = {}

        if 'no_system' in kwargs:
            params['excludeSystem'] = kwargs['no_system']

        try:
            return self.api.get(*name, params=params)

        except exc.CollectionNotFound:
            return None


class DocumentsMixin(object):

    def create(self, collection, doc, createCollection=None):
        """Create a document."""

        params = {'collection': collection}

        if createCollection is not None:
            params['createCollection'] = createCollection

        return self.api.post(json=doc, params=params)

    def get(self, *handle, **kwargs):
        """Get a document or all documents.

        :param handle: the document handle or the collection name

        """

        params = {}

        if len(handle) == 1 and '/' not in handle[0]:
            params = dict(collection=handle[0])

            # default to document handle
            params['type'] = kwargs.get('type', 'id')

            handle = ()

        return self.api.get(*handle, params=params)

    def delete(self, *handle):
        """Delete a document."""

        return self.api.delete(*handle)

    def replace(self, doc, *handle, **kwargs):
        """Replace a document."""

        return self.api.put(*handle, json=doc)

    def update(self, doc, *handle, **kwargs):
        """Partially update a document."""

        params = {}

        params['keepNull'] = kwargs.get('keep', False)
        params['mergeObjects'] = kwargs.get('merge', True)

        return self.api.patch(*handle, json=doc, params=params)


class Documents(Api, DocumentsMixin):
    pass


class Edges(Api, DocumentsMixin):
    """Edge stuff."""

    def create(self, collection, _from, _to, edge):
        params = {
            'from': str(_from),
            'to': str(_to),
            'collection': collection
        }

        return self.api.post(json=edge, params=params)


class Graphs(Api):

    def get(self, *name):
        result = self.api.get(*name)

        if name:
            return result['graph']

        return result['graphs']

    def create(self, definition):
        return self.api.post(json=definition)

    def drop(self, name):
        return self.api.delete(name)

    def vertex(self, name):
        return self.api.get(name, "vertex")['collections']


def remap_fields(dct, *include, **mapping):
    """Remap certain fields of a dict by yielding (key, value)."""

    for k, v in dct.iteritems():
        # just define all possible keys
        # this is to prevent wrong ones
        if include and k not in include:
            continue

        yield mapping.get(k, k), v


class Cursors(Api):

    """
    see https://github.com/arangodb/arangodb/issues/1285

    no underscore in query bind var
    """

    def create(self, query, bind=None, **kwargs):
        # https://docs.arangodb.com/HttpAqlQueryCursor/AccessingCursors.html
        body = dict(
            query=query,
            bindVars=bind or {}
        )

        # optional fields
        body.update(
            remap_fields(
                kwargs,
                'batch', 'ttl', 'count',
                batch='batchSize'
            )
        )

        return self.api.post(json=body)

    def pursue(self, cursor_id):
        """Just continue to load a batch from a previous call."""

        return self.api.put(cursor_id)

    def delete(self, name):
        return self.api.delete(name)


class Indexes(Api):
    def get(self, *handle, **kwargs):
        """Get a document or all documents.

        :param handle: the document handle or the collection name

        """

        params = {}

        if len(handle) == 1 and '/' not in handle[0]:
            params = dict(collection=handle[0])

            # default to document handle
            params['type'] = kwargs.get('type', 'id')

            handle = ()

        return self.api.get(*handle, params=params)

    def create(self, collection, index_type, fields=None, **kwargs):
        """Create an index."""

        params = {'collection': collection}

        doc = {
            "type": index_type,
            "fields": fields is not None and fields or []
        }

        doc.update(kwargs)

        return self.api.post(json=doc, params=params)


class Queries(Api):
    def parse(self, query):
        """Parse a query and validate it by the server."""

        return self.api.post(
            json={
                'query': query
            }
        )



class classproperty(object):
    """Create a property for the class."""

    def __init__(self, getter):
        self.getter = getter

    def __get__(self, instance, owner):
        return self.getter(owner)



"""
Some meta classes
"""

from collections import OrderedDict
from functools import wraps
from itertools import chain

from . import api, exc

import logging

LOG = logging.getLogger(__name__)


class MetaBase(type):

    """Collects and registers all classes, which need a session resp. an api.

    Provides a session property to the class, for that arangodb server connection.
    Provides also an api property mapping to the specific arango server url.
    """

    # global client factory
    __client_factory__ = None

    @classmethod
    def _set_global_client_factory(mcs, factory):
        """Set the global client factory."""

        # we take the method wrapper here
        mcs.__client_factory__ = factory.__get__

    def _set_client_factory(cls, factory):
        """Set the local client factory.
        This will only be available to subclasses of this class.
        """

        # we take the method wrapper here
        cls.__client_factory__ = factory.__get__

    @property
    def client(cls):
        """Resolves the client for this class."""

        if callable(cls.__client_factory__):
            return cls.__client_factory__(cls)()

        # just the default
        return api.SystemClient()

    @property
    def api(cls):
        raise NotImplementedError("There is no api implemented for the general meta class!")


class MetaDocumentBase(MetaBase):

    """Document type.

    Edges are also documents, but a little bit special...
    """

    # the collection name may deviate from class name
    __collection_name__ = None

    __document_base__ = None
    __documents__ = OrderedDict()

    def __init__(cls, name, bases, dct):
        super(MetaDocumentBase, cls).__init__(name, bases, dct)

        # stores an optional de/serializer
        cls.__objective__ = None

    def __new__(mcs, name, bases, dct):

        # we set our collection name to class name if not already done
        if '__collection_name__' not in dct:
            dct['__collection_name__'] = name

        cls = type.__new__(mcs, name, bases, dct)

        # register our base class or a document class
        mcs._register_class(cls)

        return cls

    @classmethod
    def _register_class(mcs, cls):
        if mcs.__document_base__ is None:
            mcs.__document_base__ = cls

        elif cls.__collection_name__ in mcs.__documents__:
            raise TypeError(
                "A collection with the same name `{0}`"
                " is already registered by {1}".format(
                    cls.__collection_name__,
                    mcs.__documents__[cls.__collection_name__]
                )
            )

        else:
            mcs.__documents__[cls.__collection_name__] = cls

    def _create_collection(cls):
        """Try to create a collection."""

        col_type = api.EDGE_COLLECTION if isinstance(cls, MetaEdgeBase) else api.DOCUMENT_COLLECTION

        col = cls.client.collections.get(cls.__collection_name__)
        if col is None:
            cls.client.collections.create(cls.__collection_name__, type=col_type)
            LOG.info("Created collection: %s", cls)

        else:
            # check if type is good
            if col['type'] != col_type:
                raise exc.ArangoException(
                    "An existing collection has the wrong type, solve this manually!",
                    col, cls)

        LOG.info("Collection in use: %s", cls)

    def _polymorph(cls, dct):
        """Just create a proper instance for this dct."""

        # dct must have an _id to infer collection
        if '_id' not in dct or '/' not in dct['_id']:
            raise TypeError("Cannot infer document type!", dct)

        collection, _ = dct['_id'].split('/')

        # lookup type
        if collection not in cls.__documents__:
            raise TypeError("Unknown document type!", collection)

        return cls.__documents__[collection](dct)

    @property
    def api(cls):
        return cls.client.documents


class MetaEdgeBase(MetaDocumentBase):

    """The edge type."""

    __edge_base__ = None
    __edges__ = OrderedDict()

    @classmethod
    def _register_class(mcs, cls):
        """Register edge base and classes."""

        if mcs.__edge_base__ is None:
            mcs.__edge_base__ = cls

        else:
            super(MetaEdgeBase, mcs)._register_class(cls)
            mcs.__edges__[cls.__collection_name__] = cls

    @property
    def api(cls):
        return cls.client.edges


class MetaCursorBase(MetaBase):

    @property
    def api(cls):
        return cls.client.cursors


class MetaIndexBase(MetaBase):

    @property
    def api(cls):
        return cls.client.indexes


class MetaQueryBase(MetaBase):

    @property
    def api(cls):
        return cls.client.queries


def find_collection_name(bases, document_cls):
    for base in bases:
        if issubclass(base, document_cls):
            # add collection name from "real" edge
            return base.__collection_name__

    raise TypeError("You have to specify an Edge base class!")


# we have to subclass from MetaEdgeBase
class MetaGraphEdge(MetaEdgeBase):
    __graph_edge_base__ = None
    __graph_edges__ = {}

    __from_vertices__ = []
    __to_vertices__ = []

    def __init__(cls, name, bases, dct):
        super(MetaGraphEdge, cls).__init__(name, bases, dct)

        # init vertices
        cls.__from_vertices__ = []
        cls.__to_vertices__ = []

        # resolve collection name
        if cls != cls.__graph_edge_base__:
            cls.__collection_name__ = find_collection_name(bases, EdgeBase)

    @classmethod
    def _register_class(mcs, cls):
        if mcs.__graph_edge_base__ is None:
            mcs.__graph_edge_base__ = cls

        else:
            mcs.__graph_edges__[cls.__collection_name__] = cls

    @property
    def graph_api(cls):
        client = cls.client

        return api.Edges(client.api(client.database, 'gharial', cls.__graph__.__graph_name__, 'edge'))

    def from_vertex(cls, vertex):
        if vertex not in cls.__from_vertices__:
            cls.__from_vertices__.append(vertex)

        return vertex

    def to_vertex(cls, vertex):
        if vertex not in cls.__to_vertices__:
            cls.__to_vertices__.append(vertex)

        return vertex

    @property
    def __definition__(cls):
        return {
            "collection": cls.__collection_name__,
            "from": [col.__collection_name__ for col in cls.__from_vertices__],
            "to": [col.__collection_name__ for col in cls.__to_vertices__]
        }


class MetaGraphVertex(MetaDocumentBase):
    __graph_vertex_base__ = None
    __graph_vertices__ = {}

    def __init__(cls, name, bases, dct):
        super(MetaGraphVertex, cls).__init__(name, bases, dct)

        # resolve collection name
        if cls != cls.__graph_vertex_base__:
            cls.__collection_name__ = find_collection_name(bases, DocumentBase)

    @classmethod
    def _register_class(mcs, cls):
        if mcs.__graph_vertex_base__ is None:
            mcs.__graph_vertex_base__ = cls

        else:
            mcs.__graph_vertices__[cls.__collection_name__] = cls

    @property
    def graph_api(cls):
        client = cls.client

        return api.Edges(client.api(client.database, 'gharial', cls.__graph__.__graph_name__, 'vertex'))


class MetaGraph(MetaBase):
    __graph_base__ = None
    __graphs__ = {}

    __graph_name__ = None

    def __new__(mcs, name, bases, dct):

        cls = type.__new__(mcs, name, bases, dct)

        if mcs.__graph_base__ is None:
            mcs.__graph_base__ = cls

        else:
            mcs.__graphs__[cls.__name__] = cls

        return cls

    def __init__(cls, name, bases, dct):
        super(MetaGraph, cls).__init__(name, bases, dct)

        cls.__graph_edges__ = {}
        cls.__graph_vertices__ = {}

        # graph name is always the first none base
        if cls.__graph_base__ is not cls and cls.__graph_name__ is None:
            cls.__graph_name__ = name

        # lookup graph classes to attach graph specific api proxy
        for name in dir(cls):

            attr = getattr(cls, name)
            if isinstance(attr, MetaGraphEdge) or isinstance(attr, MetaGraphVertex):
                attr.__graph__ = cls

                # collect graph edges and vertices
                if isinstance(attr, MetaGraphEdge):
                    cls.__graph_edges__[attr.__collection_name__] = attr

                else:
                    cls.__graph_vertices__[attr.__collection_name__] = attr

    @property
    def __definition__(cls):
        """ArangoDB graph definition."""

        # orphans are not in an edge definition
        in_edge = set()

        edge_definitions = []
        for edge in cls.__graph_edges__.itervalues():
            edge_definitions.append(edge.__definition__)
            in_edge.update(set(edge.__from_vertices__) | set(edge.__to_vertices__))

        orphans = [
            name
            for name, vertice in cls.__graph_vertices__.iteritems()
            if vertice not in in_edge
        ]

        return {
            'name': cls.__graph_name__,
            'edgeDefinitions': edge_definitions,
            'orphanCollections': orphans
        }

    @property
    def api(cls):
        return cls.client.graphs

    def _create_graph(cls):
        # first create all graph collections
        for edge in chain(
                cls.__graph_edges__.itervalues(),
                cls.__graph_vertices__.itervalues()
        ):
            edge._create_collection()       # pylint: disable=W0212

        # check for existing graph
        try:
            definition = cls.api.get(cls.__graph_name__)

            if definition != cls.__definition__:
                # TODO replace old with new definition
                pass

        except exc.GraphNotFound:
            # post graph definition
            return cls.api.create(cls.__definition__)


def polymorph(wrapped):
    """Decorate a function which returns a raw ArangoDB document to create a document class instance."""

    # TODO
    # polymorph should deal with graph context
    # so that a document gets becomes a vertex

    @wraps(wrapped)
    def decorator(*args, **kwargs):
        doc = wrapped(*args, **kwargs)

        return BaseDocument._polymorph(doc)         # pylint: disable=W0212

    return decorator


class BaseDocument(object):

    """Is an object, which is able to connect to a session."""

    __metaclass__ = MetaDocumentBase

    def __init__(self, *args, **kwargs):
        self.__data__ = {}

        # we have to call __setitem__
        self.update(*args, **kwargs)

    def __setitem__(self, key, value):
        self.__data__[key] = value

    def __getitem__(self, key):
        return self.__data__[key]

    def __delitem__(self, key):
        del self.__data__[key]

    def __contains__(self, key):
        return key in self.__data__

    def __iter__(self):
        return self.__data__.__iter__()

    def keys(self):
        return self.__data__.keys()

    def iteritems(self):
        return self.__data__.iteritems()

    def iterkeys(self):
        return self.__data__.iterkeys()

    def itervalues(self):
        return self.__data__.itervalues()

    def copy(self):
        clone = self.__class__()
        clone.__data__ = self.__data__.copy()
        return clone

    def update(self, *args, **kwargs):
        for key, value in dict(*args, **kwargs).iteritems():
            self[key] = value

    def get(self, key, default=None):
        if key in self:
            return self[key]

        return default

    @property
    def _id(self):
        return self['_id']

    @_id.setter
    def _id(self, value):
        self['_id'] = value

    @property
    def _key(self):
        return self['_key']

    @_key.setter
    def _key(self, value):
        self['_key'] = value

    @property
    def _rev(self):
        return self['_rev']

    @_rev.setter
    def _rev(self, value):
        self['_rev'] = value

    @classmethod
    def _iter_all(cls):
        """query all documents and create instances.

        This can be very big, use this just for debugging!
        """
        for _id in cls.api.get(cls.__collection_name__)['documents']:
            yield cls.load(_id)

    @classmethod
    def _create(cls, doc):
        """Create a db instance on the server."""

        return cls.api.create(cls.__collection_name__, doc)

    @classmethod
    def deserialize(cls, doc):
        """Take the deserializer to validate the document."""

        if cls.__objective__ is not None:
            return cls.__objective__.deserialize(doc)

        return doc.copy()

    def serialize(self):
        """Take the serializer to adjust the document."""

        if self.__objective__ is not None:
            return self.__objective__.serialize(self.__data__)

        return self.__data__.copy()

    @classmethod
    @polymorph
    def load(cls, key):
        """Just create a fresh instance by requesting the document by key."""

        if '/' in key:
            name, key = key.split('/')

        else:
            name = cls.__collection_name__

        doc = cls.api.get(name, key)

        return cls.deserialize(doc)

    def save(self):
        """Save the document to db or update."""

        serialized = self.serialize()

        # test for existing key
        if '_id' in self:
            # replace
            doc = self.__class__.api.replace(serialized, self['_id'])

        else:
            # create
            doc = self._create(serialized)

        # update self
        for k in ('_id', '_key', '_rev'):
            self[k] = doc[k]

    def delete(self):
        """Delete a document."""

        return self.__class__.api.delete(self['_id'])

    def __str__(self):
        return self.get(
            '_id',
            '/'.join(
                (self.__class__.__collection_name__, self.get('key', ''))
            )
        )


class DocumentBase(BaseDocument):

    """Just an arangodb document."""

    __metaclass__ = MetaDocumentBase


class EdgeBase(BaseDocument):

    __metaclass__ = MetaEdgeBase


class CursorBase(object):

    __metaclass__ = MetaCursorBase


class IndexBase(object):

    """An index representation."""

    __metaclass__ = MetaIndexBase


class GraphBase(object):

    """The base class for edges and vertices."""

    __graph__ = None

    @classmethod
    def _create(cls, doc):
        """Create a db instance on the server."""

        return cls.graph_api.create(cls.__collection_name__, doc)

    def save(self):
        """Save the document to db or update."""

        serialized = self.serialize()

        # test for existing key
        if '_id' in self:
            # replace
            doc = self.__class__.graph_api.replace(serialized, self['_id'])

        else:
            # create
            doc = self._create(serialized)

        # update self
        for k in ('_id', '_key', '_rev'):
            self[k] = doc[k]

    def delete(self):
        """Delete a document."""

        return self.__class__.graph_api.delete(self['_id'])



"""
The arangodb errors.
"""


class ArangoException(Exception):
    pass


class ContentTypeException(ArangoException):
    pass


class MetaApiError(type):
    """Just for polymorphic instantiation."""

    base = None
    classes = {}

    def __new__(mcs, name, bases, dct):
        cls = type.__new__(mcs, name, bases, dct)

        if mcs.base is None:
            mcs.base = cls

        else:
            error_num = dct.get('error_num')
            if error_num is not None:

                # this should not happen
                if error_num in mcs.classes:
                    raise ArangoException("You defined an arango error ({}) twice!".format(error_num),
                                          error_num, mcs.classes[error_num], cls)

                mcs.classes[error_num] = cls

        return cls

    def __call__(cls, code, num, message, *args, **kwargs):
        # polymorphic part
        # lookup exception type
        if num in cls.classes:
            cls = cls.classes[num]

        return type.__call__(cls, code, num, message, *args, **kwargs)


class ApiError(ArangoException):
    """Raise when an api error occurs."""

    __metaclass__ = MetaApiError

    error_num = None

    def __init__(self, code, num, message, func=None, args=None, kwargs=None):
        self.code = code
        self.num = num
        self.message = message

        self.func = func
        self.args = args
        self.kwargs = kwargs

        super(ApiError, self).__init__(code, num, message, func, args, kwargs)


# basically an implementation of https://docs.arangodb.com/ErrorCodes/README.html

class CollectionNotFound(ApiError):
    error_num = 1203


class CollectionTypeInvalid(ApiError):
    error_num = 1218


class DatabaseNotFound(ApiError):
    error_num = 1228


class UniqueConstraintViolated(ApiError):
    error_num = 1210


class GraphError(ApiError):
    pass


class EdgeCollectionAlreadyUsed(GraphError):
    error_num = 1921


class GraphNotFound(GraphError):
    error_num = 1924



import inspect
from itertools import chain

from . import db, api, exc, meta


import logging
logging.basicConfig(level=logging.DEBUG)


class GraphEdge(meta.GraphBase):
    __metaclass__ = meta.MetaGraphEdge


class GraphVertex(meta.GraphBase):
    __metaclass__ = meta.MetaGraphVertex


class Graph(object):
    __metaclass__ = meta.MetaGraph



"""Some classes to easy work with arangodb."""

from . import meta, util, query

import logging

LOG = logging.getLogger(__name__)


class QueryMixin(object):
    # pylint: disable=E0213

    @util.classproperty
    def alias(cls):
        """A query alias for this collection."""

        return query.Alias(cls.__collection_name__)

    @util.classproperty
    def query(cls):
        """Prepare a query against this collection.

        The default action is to return the alias.
        """

        return query.Query(cls.alias, query.Collection(cls)).action(cls.alias)


class Document(meta.DocumentBase, QueryMixin):
    pass


EDGE_DIRECTION_ANY = 'any'
EDGE_DIRECTION_INBOUND = 'inbound'
EDGE_DIRECTION_OUTBOUND = 'outbound'


class Edge(meta.EdgeBase, QueryMixin):

    """An edge between two documents.

    When the edge is loaded the two dowuments are also loaded.
    """

    def __init__(self, *args, **kwargs):
        """
        call scheme:

            Edge([_from, _to,] [iterable,] **kwargs)

        If _from and _to are not given, they have to be set later before saving!
        """

        # split args
        args_len = len(args)

        if args_len < 2:
            # _to and _from must be in kwargs or args
            pass

        else:
            kwargs['_from'], kwargs['_to'] = args[:2]
            args = args[2:]

        # else:
            # pass

        super(Edge, self).__init__(*args, **kwargs)

    def __setitem__(self, key, value):
        if key in ('_from', '_to') and isinstance(value, meta.BaseDocument):
            # check for documents and reduce to _id
            if '_id' not in value:
                raise TypeError(
                    "The document for setting `{0}` has no `_id`: {1}"
                    .format(key, value)
                )

            super(Edge, self).__setitem__(key, value['_id'])

        else:
            super(Edge, self).__setitem__(key, value)

    @property
    def _from(self):
        return self['_from']

    @_from.setter
    def _from(self, value):
        self['_from'] = value

    @property
    def _to(self):
        return self['_to']

    @_to.setter
    def _to(self, value):
        self['_to'] = value

    @classmethod
    def _create(cls, doc):
        """Create a db instance."""

        assert '_from' in doc and '_to' in doc, \
            "You must create an edge ether by calling __init__ " \
            "with _from and _to args or an appropriate dict!"

        return cls.api.create(
            cls.__collection_name__,
            doc['_from'],
            doc['_to'],
            doc)

    @classmethod
    def load(cls, key):
        """Load the edge and connected documents."""

        edge = super(Edge, cls).load(key)

        edge['_from'] = Document.load(edge['_from'])
        edge['_to'] = Document.load(edge['_to'])

        return edge

    @classmethod
    def connections_query(cls, alias, document, direction=EDGE_DIRECTION_ANY):
        assert isinstance(document, meta.BaseDocument), "document is Document or Edge"

        # pylint: disable=W0212
        q = query.Query(
            alias,
            query.PATHS(
                query.Collection(document),
                query.Collection(cls),
                direction)
        )\
            .filter(alias.source._id == document._id)\
            .filter(query.LENGTH(alias.edges) == 1)\
            .action(alias.destination)

        return q

    @classmethod
    def connections(cls, document, collection=None, direction=EDGE_DIRECTION_ANY):
        alias = query.Alias('p')

        q = cls.connections_query(alias, document, direction)

        if collection is not None:
            # pylint: disable=W0212
            q = q\
                .filter(query.FIND_FIRST(alias.destination._id, collection.__collection_name__))

        return q.cursor.iter_documents()

    @classmethod
    def inbounds(cls, document, collection=None):
        return cls.connections(document, collection, direction=EDGE_DIRECTION_INBOUND)

    @classmethod
    def outbounds(cls, document, collection=None):
        return cls.connections(document, collection, direction=EDGE_DIRECTION_OUTBOUND)


class Index(meta.IndexBase):

    """An index representation."""

    collection = None
    index_type = None
    unique = False


INDEX_TYPE_HASH = "hash"


class Hash(Index):

    """A hash index."""

    collection = None
    index_type = INDEX_TYPE_HASH

    # unique can be setted at class level
    unique = False

    def __init__(self, *fields, **kwargs):
        self.fields = fields

        if "unique" in kwargs:
            self.unique = kwargs['unique']

        if "collection" in kwargs:
            self.collection = kwargs['collection']

        if self.collection is None:
            raise TypeError("No index collection specified!")

    def save(self):
        if isinstance(self.collection, meta.BaseDocument):
            collection = self.collection.__collection_name__

        else:
            collection = self.collection

        result = self.api.create(collection, self.index_type, fields=self.fields, unique=self.unique)

        return result


class UniqueHash(Hash):

    """A unique hash index."""

    unique = True



"""package setup"""

__version__ = "0.0.12"

import sys, os

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    """Our test runner."""

    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ["tests"]

    def finalize_options(self):
        # pylint: disable=W0201
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


def read(*paths):
    """Build a file path from *paths* and return the contents."""
    with open(os.path.join(*paths), 'r') as f:
        return f.read()


setup(
    name="arangodb",
    author="Oliver Berger",
    author_email="diefans@gmail.com",
    url="https://github.com/diefans/python-arangodb",
    description='A Python client for ArangoDB REST API.',
    long_description=read('README.rst'),
    version=__version__,
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2.7',
        'Topic :: Database',
    ],
    license='Apache License Version 2.0',

    keywords="ArangoDB REST API Requests graph nosql database AQL",

    package_dir={'': 'src'},
    namespace_packages=['arangodb'],
    packages=find_packages(
        'src',
        exclude=["tests*"]
    ),
    entry_points="",

    install_requires=[
        'requests',
    ],

    cmdclass={'test': PyTest},
    tests_require=[
        # tests
        'pytest',
        'pytest-pep8',
        'pytest-cache',
        'pytest-random',
        'mock',
    ]
)



from . import meta

import logging

LOG = logging.getLogger(__name__)


class Cursor(meta.CursorBase):

    """A cursor is created to perform queries."""

    def __init__(self, query, bind=None, **kwargs):
        self.query = query
        self.bind = bind
        self.kwargs = kwargs

    def iter_result(self):
        """Iterate over all batches of result."""

        LOG.debug("Create cursor: `%s`, %s, %s", self.query, self.bind, self.kwargs)
        cursor = self.__class__.api.create(self.query, bind=self.bind, **self.kwargs)

        while cursor['result']:
            for result in cursor['result']:
                yield result

            if not cursor['hasMore']:
                # step out
                break

            # fetch next batch
            cursor = self.__class__.api.pursue(cursor['id'])

    def iter_documents(self):

        """If you expect document instances to be returned from the cursor,
        call this to instantiate them."""

        for doc in self.iter_result():
            yield meta.BaseDocument._polymorph(doc)         # pylint: disable=W0212

    def first_document(self):
        for doc in self.iter_documents():
            return doc
